
def get_classification_based_on_time_range(start_time, end_time, labels):
    for label in labels:
        if start_time >= label.start_seconds and end_time < label.end_seconds:
            return label.label
    raise Exception("can't find label matching start %s and end time %s" % (start_time, end_time))


def frame_level_evaluation(ground_truth_labels, detected_labels, frame_sec=0.01, verbose=False):

    correctly_classified = 0
    total_frames = 0
    incorrectly_classified = 0

    for ground_truth_label in ground_truth_labels:
        current = ground_truth_label.start_seconds
        while current < ground_truth_label.end_seconds - frame_sec:
            total_frames += 1
            try:
                if ground_truth_label.label == get_classification_based_on_time_range(current, current+frame_sec,
                                                                                      detected_labels):
                    correctly_classified += 1
                else:
                    incorrectly_classified += 1

            except Exception as e:
                pass

            current += frame_sec

    if verbose:
        print(correctly_classified)
        print(incorrectly_classified)
        print(total_frames)

    precision = correctly_classified / float(correctly_classified + incorrectly_classified)
    recall = correctly_classified / float(total_frames)
    fs = (2 * precision * recall) / (precision + recall)

    if verbose:
        print("FRAME LEVEL EVALUATION => Precision: %s Recall: %s F1 Score: %s" % (precision, recall, fs))

    return precision, recall, fs


def segments_precision_recall_fs(ground_truth_segments, detected_segments, tolerance, verbose=False):
    if verbose:
        print("Ground truth segments: %s" % len(ground_truth_segments))
        print("Detected segments: %s" % len(detected_segments))

    correctly_detected_transition_points = 0

    for i in range(0, len(ground_truth_segments)):
        ground_segment = ground_truth_segments[i]
        for detected_segment in detected_segments:
            if abs(detected_segment.start_seconds - ground_segment.start_seconds) <= tolerance \
                    and abs(detected_segment.end_seconds - ground_segment.end_seconds) <= tolerance \
                    and detected_segment.label == ground_segment.label:
                        correctly_detected_transition_points += 1
                        break

    if verbose:
        print("Correctly detected segments: %s" % correctly_detected_transition_points)

    if len(detected_segments) > 0:
        precision = correctly_detected_transition_points / float(len(detected_segments)-1)
    else:
        precision = 0
    if len(ground_truth_segments) > 0:
        recall = correctly_detected_transition_points / float(len(ground_truth_segments)-1)
    else:
        recall = 0
    if precision + recall != 0:
        fs = (2 * precision * recall) / (precision + recall)
    else:
        fs = 0
    if verbose:
        print("SEGMENT LEVEL EVALUATION => Precision: %s Recall: %s F1 Score: %s" % (precision, recall, fs))

    return precision, recall, fs


def transition_points_precision_recall_fs(ground_truth_segments, detected_segments, tolerance, verbose=False):
    if verbose:
        print("Ground truth segments: %s" % len(ground_truth_segments))
        print("Detected segments: %s" % len(detected_segments))

    correctly_detected_transition_points = 0

    for i in range(0, len(ground_truth_segments)):
        ground_segment = ground_truth_segments[i]
        for detected_segment in detected_segments:
            if abs(detected_segment.start_seconds - ground_segment.end_seconds) <= tolerance:
                correctly_detected_transition_points += 1
                break

    if verbose:
        print("Correctly detected transition points: %s" % correctly_detected_transition_points)

    if len(detected_segments) > 0:
        precision = correctly_detected_transition_points / float(len(detected_segments)-1)
    else:
        precision = 0
    if len(ground_truth_segments) > 0:
        recall = correctly_detected_transition_points / float(len(ground_truth_segments)-1)
    else:
        recall = 0
    if precision + recall != 0:
        fs = (2 * precision * recall) / (precision + recall)
    else:
        fs = 0

    if verbose:
        print("SEGMENT LEVEL EVALUATION => Precision: %s Recall: %s F1 Score: %s" % (precision, recall, fs))

    return precision, recall, fs


import math

from matplotlib.patches import Rectangle
from sac.model.audacity_label import AudacityLabel
from scipy.ndimage import uniform_filter
import numpy as np
from scipy.signal import find_peaks_cwt
import matplotlib.pyplot as plt
import peakutils


def get_segments(timestamps, sm, filter_width, kernel_width, thresh=0.25, draw=False):

    sm = uniform_filter(sm, filter_width)

    peaks, convolution_values = checkerboard_matrix_filtering(sm, kernel_width, thresh=thresh)

    if draw:
        fig = plt.figure()
        ax = fig.add_subplot(111)
        for i in range(0, len(peaks)-1):
            ax.add_patch(Rectangle((peaks[i], peaks[i]), peaks[i+1] - peaks[i], peaks[i+1] - peaks[i],
                                   fill=False, edgecolor='b'))

    segments = calculate_segment_start_end_times_from_peak_positions(peaks, timestamps)

    if draw:
        ax.imshow(sm, cmap=plt.cm.gray)
        plt.show()

    return segments


def calculate_segment_start_end_times_from_peak_positions(peaks, timestamps):
    window_in_seconds = get_window_in_seconds(timestamps)
    segments = []
    for i in range(0, len(peaks) - 1):
        segments.append(AudacityLabel(peaks[i] * window_in_seconds, peaks[i + 1] * window_in_seconds, "-"))

    return segments


def get_window_in_seconds(timestamps):
    return timestamps[1] - timestamps[0]


def get_gaussian_kernel(kernel_width, gaussian_param):

    quarter = np.ones((kernel_width, kernel_width))

    for i in range(0, quarter.shape[0]):
        for j in range(0, quarter.shape[1]):
            m = max([i, j])
            quarter[i, j] = math.exp(gaussian_param * (-m**2))

    return np.vstack((
        np.hstack((
            -1 * np.flipud(np.fliplr(quarter)), np.flipud(quarter)
        )),
        np.hstack((
            np.fliplr(quarter), -1 * quarter)
        ))
    )


def get_checkerboard_matrix(kernel_width, kernel_type="default", gaussian_param=0.1):

    """
    example matrix for width = 2

    -1  -1    1   1
    -1  -1    1   1
     1   1   -1  -1
     1   1   -1  -1

    :param kernel_type:
    :param kernel_width:
    :return:
    """

    if kernel_type is "gaussian":
        return get_gaussian_kernel(kernel_width, gaussian_param)

    if kernel_type is "default":
        return np.vstack((
            np.hstack((
                -1 * np.ones((kernel_width, kernel_width)), np.ones((kernel_width, kernel_width))
            )),
            np.hstack((
                np.ones((kernel_width, kernel_width)), -1 * np.ones((kernel_width, kernel_width))
            ))
        ))

    elif kernel_type is "bottom_right":
        return np.vstack((
            np.hstack((
                np.ones((kernel_width, kernel_width)), np.ones((kernel_width, kernel_width))
            )),
            np.hstack((
                np.ones((kernel_width, kernel_width)), -1 * np.ones((kernel_width, kernel_width))
            ))
        ))

    elif kernel_type is "top_left":
        return np.vstack((
            np.hstack((
                -1 * np.ones((kernel_width, kernel_width)), np.ones((kernel_width, kernel_width))
            )),
            np.hstack((
                np.ones((kernel_width, kernel_width)), np.ones((kernel_width, kernel_width))
            ))
        ))


def checkerboard_matrix_filtering(similarity_matrix, kernel_width, kernel_type="default", thresh=0.25):

    """
    Moving the checkerboard matrix over the main diagonal of the similarity matrix one sample at a time.

    :param kernel_type:
    :param thresh:
    :param similarity_matrix:
    :param kernel_width: the size of one quarter of the checkerboard matrix
    :return: peaks and convolution values
    """

    checkerboard_matrix = get_checkerboard_matrix(kernel_width, kernel_type)

    # The values calculated in this step are starting from the 'kernel_width' position and ending
    # at length - kernel_width
    d = []
    for i in range(0, similarity_matrix.shape[0] - 2 * kernel_width):
        base = similarity_matrix[i:i + kernel_width * 2, i:i + kernel_width * 2]
        d.append(np.sum(np.multiply(base, checkerboard_matrix)))

    # The missing values from 0 to kernel_width are calculated here
    top_left_d = []
    for i in range(0, kernel_width):
        base = similarity_matrix[0:i + kernel_width, 0:i + kernel_width]
        top_left_d.append(np.sum(np.multiply(base, checkerboard_matrix[kernel_width - i:, kernel_width - i:])))

    # The missing kernel_width values at the bottom right are set to 0
    convolution_values = top_left_d + d + [0 for i in range(0, kernel_width)]

    # peaks = find_peaks_cwt(convolution_values, np.arange(1, peak_range))
    peaks = peakutils.indexes(convolution_values, thres=thresh)

    peaks = [0] + list(peaks) + [len(convolution_values)-1]
    return peaks, convolution_values



import pickle
from unittest import TestCase

import numpy as np
import os
from sac.model.audacity_label import AudacityLabel
from sac.util import Util


class UtilTests(TestCase):
    def setUp(self):
        self.stats_csv = os.path.join(os.path.dirname(__file__), 'fixtures/example_yaafe_stats.csv')
        self.no_stats_csv = os.path.join(os.path.dirname(__file__), 'fixtures/example_yaafe_no_stats.csv')
        self.double_stats_csv = os.path.join(os.path.dirname(__file__), 'fixtures/example_yaafe_double_stats.csv')
        self.no_stats_derivate = os.path.join(os.path.dirname(__file__), 'fixtures/example_yaafe_no_stats_derivate.csv')
        self.stats_derivate = os.path.join(os.path.dirname(__file__), 'fixtures/example_yaafe_stats_derivate.csv')

        # with open(os.path.abspath(os.path.join(os.path.dirname(__file__), 'fixtures/sm.pickle')), 'rb') as f:
        #     self.sm = pickle.load(f)

        # with open(os.path.abspath(os.path.join(os.path.dirname(__file__), 'fixtures/timestamps.pickle')), 'rb') as f:
        #     self.timestamps = pickle.load(f)

    # # yaafe header parser
    def test_parse_yaafe_header_stats(self):
        header = Util.parse_yaafe_header(self.stats_csv)
        self.assertEqual(22050, header['samplerate'])
        self.assertEqual(15360, header['effective_step_size'])

    def test_parse_yaafe_header_no_stats(self):
        header = Util.parse_yaafe_header(self.no_stats_csv)
        self.assertEqual(22050, header['samplerate'])
        self.assertEqual(512, header['effective_step_size'])

    def test_parse_yaafe_header_double_stats(self):
        header = Util.parse_yaafe_header(self.double_stats_csv)
        self.assertEqual(22050, header['samplerate'])
        self.assertEqual(460800, header['effective_step_size'])

    def test_parse_yaafe_header_no_stats_derivate(self):
        header = Util.parse_yaafe_header(self.no_stats_derivate)
        self.assertEqual(22050, header['samplerate'])
        self.assertEqual(512, header['effective_step_size'])

    def test_parse_yaafe_header_stats_derivate(self):
        header = Util.parse_yaafe_header(self.stats_derivate)
        self.assertEqual(22050, header['samplerate'])
        self.assertEqual(15360, header['effective_step_size'])

    # # yaafe csv parser
    def test_load_yaafe_csv_no_stats(self):
        timestamps, features = Util.load_yaafe_csv(self.no_stats_csv)
        self.assertEqual((17, 2), features.shape)
        self.assertEqual(0.37151927437641724, timestamps[-1])

    def test_load_yaafe_csv_stats(self):
        timestamps, features = Util.load_yaafe_csv(self.stats_csv)
        self.assertEqual((17, 2), features.shape)
        self.assertEqual(11.145578231292516, timestamps[-1])

    def test_load_yaafe_csv_double_stats(self):
        timestamps, features = Util.load_yaafe_csv(self.double_stats_csv)
        self.assertEqual((17, 2), features.shape)
        self.assertEqual(334.3673469387755, timestamps[-1])

    # # merge classifications
    def test_generate_labels_from_classifications(self):
        classifications = [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1]
        timestamps = [0.0, 0.6965986394557823, 1.3931972789115645, 2.089795918367347, 2.786394557823129,
                      3.4829931972789114, 4.179591836734694, 4.876190476190477, 5.572789115646258, 6.2693877551020405,
                      6.965986394557823, 7.662585034013605, 8.359183673469389, 9.05578231292517, 9.752380952380953,
                      10.448979591836734, 11.145578231292516, 11.842176870748299, 12.538775510204081,
                      13.235374149659863, 13.931972789115646, 14.628571428571428, 15.32517006802721, 16.021768707482995]

        labels = Util.generate_labels_from_classifications(classifications, timestamps)

        expected_labels = [AudacityLabel(0.0, 1.3931972789115645, 1),
                           AudacityLabel(1.3931972789115645, 9.752380952380953, 0),
                           AudacityLabel(9.752380952380953, 10.448979591836736, 1),
                           AudacityLabel(10.448979591836734, 11.145578231292516, 0),
                           AudacityLabel(11.145578231292516, 12.538775510204081, 1),
                           AudacityLabel(12.538775510204081, 13.235374149659863, 0),
                           AudacityLabel(13.235374149659863, 16.718367346938777, 1)]

        self.assertListEqual(expected_labels, labels)

    def test_combine_adjacent_labels_of_the_same_class(self):
        input_labels = [
            AudacityLabel(0, 10, "m"),
            AudacityLabel(10, 20, "m"),
            AudacityLabel(20, 21, "s"),
            AudacityLabel(21, 22, "s"),
            AudacityLabel(22, 23, "s"),
            AudacityLabel(23, 30, "m")
        ]
        expected_labels = [
            AudacityLabel(0, 20, "m"),
            AudacityLabel(20, 23, "s"),
            AudacityLabel(23, 30, "m"),
        ]

        actual_labels = Util.combine_adjacent_labels_of_the_same_class(input_labels)

        self.assertListEqual(expected_labels, actual_labels)

    def test_calculate_classes_percentages(self):
        classifications = [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1]
        percentages = Util.calculate_classes_percentages(classifications)

        self.assertAlmostEqual(0.5833333333333334, percentages[0])
        self.assertAlmostEqual(0.4166666666666667, percentages[1])

    def test_get_annotated_data(self):
        timestamps = [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0]
        data = np.random.rand(7, 10)
        labels = [
            AudacityLabel(1.2, 2.5, "m"),
            AudacityLabel(4.5, 6.0, "s")
        ]

        annotated_data = Util.get_annotated_data(timestamps, data, labels)
        self.assertTrue("m" in annotated_data)
        self.assertTrue("s" in annotated_data)
        self.assertTrue(data[2, :] in annotated_data["m"])
        self.assertTrue(data[5, :] in annotated_data["s"])
        self.assertTrue(data[6, :] in annotated_data["s"])

    def test_get_annotated_data_x_y(self):
        timestamps = [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]
        data = np.random.rand(8, 10)
        labels = [
            AudacityLabel(1.2, 2.5, "m"),
            AudacityLabel(4.5, 6.2, "s")
        ]

        x, y, classes, timestamps = Util.get_annotated_data_x_y(timestamps, data, labels)

        self.assertEqual(3, x.shape[0])
        self.assertListEqual(["m", "s", "s"], y)
        self.assertListEqual(["m", "s"], classes)
        self.assertListEqual([0.0, 1.0, 2.0], timestamps)

        labels = [
            AudacityLabel(1.0, 5.5, 'A'),
            AudacityLabel(5.5, 10.0, 'B'),
            AudacityLabel(15.0, 20.5, 'C')
        ]

        X, y, classes, new_timestamps = Util.get_annotated_data_x_y([float(i) for i in range(0, 25)], np.ones((25, 10)),
                                                                    labels)

        self.assertListEqual(['A', 'A', 'A', 'A', 'A', 'B', 'B', 'B', 'B', 'C', 'C', 'C', 'C', 'C', 'C'], y)
        self.assertListEqual([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0, 14.0], new_timestamps)


    def test_split_data_based_on_annotation(self):
        X = np.array([
            [1, 2, 3, 4],
            [2, 3, 4, 5],
            [4, 5, 6, 7]
        ])
        Y = [0, 0, 1]
        classes = ["music", "speech"]

        data = Util.split_data_based_on_annotation(X, Y, classes)

        self.assertEqual(data["music"].shape[0], 2)
        self.assertEqual(data["speech"].shape[0], 1)

    # def test_kmeans_image_quantisation(self):
    #     quantised_image = Util.kmeans_image_quantisation(self.sm, 5)
    #     self.assertEquals(5, len(np.unique(quantised_image)))

    def test_combine_peaks(self):

        a_peaks = [3, 8]
        a_peak_values = [2, 2, 2, 3, 2, 2, 2, 4, 5, 4, 2, 2, 2]

        b_peaks = [6]
        b_peak_values = [2, 2, 2, 2, 2, 2, 7, 2, 2, 2, 2, 2, 2]

        p, v = Util.combine_peaks(a_peaks, a_peak_values, b_peaks, b_peak_values)

        self.assertListEqual([3, 6, 8], p)

    def test_non_maximum_suppression(self):
        peaks = [2, 3, 4, 7, 8, 9]
        values = [2, 2, 2, 3, 2, 2, 2, 4, 5, 4, 2, 2, 2]
        p, v = Util.non_maximum_suppression(peaks, values)

        self.assertListEqual([3, 8], p)
        self.assertListEqual(values, v)

    def test_get_annotated_labels_from_predictions_and_sm_segments(self):

        timestamps = [0, 1, 2, 3, 4, 5, 6]
        segments = [
            AudacityLabel(0, 2.5, '-'),
            AudacityLabel(2.5, 5.5, '-')
        ]
        frame_level_predictions = np.array(['v', 'v', 'v', 's', 's', 'v', 'v'])

        labels = Util.get_annotated_labels_from_predictions_and_sm_segments(frame_level_predictions, segments, timestamps)

        self.assertListEqual(['v', 's'], [l.label for l in labels])

    def test_get_unshifted_labels(self):

        predicted_lbls = [
            AudacityLabel(1.0, 3.0, "A"),
            AudacityLabel(8.0, 12.0, "B")
        ]

        shifted_unshifted_labels = [
            {
                'old_label': AudacityLabel(1.0, 5.0, 'A'),
                'new_label': AudacityLabel(0.0, 4.0, 'A'),
                'shift': 1.0
            },
            {
                'old_label': AudacityLabel(5.0, 10.0, 'B'),
                'new_label': AudacityLabel(4.0, 9.0, 'B'),
                'shift': 1.0
            },
            {
                'old_label': AudacityLabel(15.0, 20.0, 'B'),
                'new_label': AudacityLabel(9.0, 14.0, 'B'),
                'shift': 6.0
            }
        ]

        lbls = Util.get_unshifted_labels(predicted_lbls, shifted_unshifted_labels)

        self.assertListEqual([AudacityLabel(2.0, 4.0, 'A'), AudacityLabel(9.0, 18.0, 'B')], lbls)

    def test_get_unshifted_timestamps(self):

        lbls = [
            {
                'old_label': AudacityLabel(1.0, 5.0, 'A'),
                'new_label': AudacityLabel(0.0, 4.0, 'A'),
                'shift': 1.0
            },
            {
                'old_label': AudacityLabel(5.0, 10.0, 'B'),
                'new_label': AudacityLabel(4.0, 9.0, 'B'),
                'shift': 1.0
            },
            {
                'old_label': AudacityLabel(15.0, 20.0, 'B'),
                'new_label': AudacityLabel(9.0, 14.0, 'B'),
                'shift': 6.0
            }
        ]

        shifted_timestamps = [3.0, 4.0, 11.0]
        expected_unshifted_timestamps = [
            shifted_timestamps[0] + lbls[0]['shift'],
            shifted_timestamps[1] + lbls[1]['shift'],
            shifted_timestamps[2] + lbls[2]['shift']
        ]
        unshifted_timestamps = Util.get_unshifted_timestamps(shifted_timestamps, lbls)
        self.assertListEqual(expected_unshifted_timestamps, unshifted_timestamps)

    def test_get_annotation_time_shift(self):

        labelsA = [
            AudacityLabel(1.0, 5.0, 'A'),
            AudacityLabel(5.0, 10.0, 'B'),
            AudacityLabel(15.0, 20.0, 'B')
        ]

        new_labels = Util.get_annotation_time_shift(labelsA)

        expected = [
                {
                    'old_label': AudacityLabel(1.0, 5.0, 'A'),
                    'new_label': AudacityLabel(0.0, 4.0, 'A'),
                    'shift': 1.0
                },
                {
                    'old_label': AudacityLabel(5.0, 10.0, 'B'),
                    'new_label': AudacityLabel(4.0, 9.0, 'B'),
                    'shift': 1.0
                },
                {
                    'old_label': AudacityLabel(15.0, 20.0, 'B'),
                    'new_label': AudacityLabel(9.0, 14.0, 'B'),
                    'shift': 6.0
                }
            ]

        self.assertListEqual(expected, new_labels)



import numpy as np


def find_peaks(series_data, sliding_window_size=20, percentage_of_sliding_window_votes=0.7):

    ranking = np.zeros(len(series_data))

    for i in range(len(series_data)):
        window_data = series_data[i:i+sliding_window_size]

        ranking[i + np.argmax(window_data)] += 1

    ranking = ranking > percentage_of_sliding_window_votes * sliding_window_size

    peaks = np.where(ranking >= 1)[0].tolist()
    return peaks



from sac.util import Util
import collections
import contextlib
import wave
import webrtcvad
from sac.model.audacity_label import AudacityLabel


class Vad(object):

    def _read_wave(self, path):
        """Reads a .wav file.
        Takes the path, and returns (PCM audio data, sample rate).
        """
        with contextlib.closing(wave.open(path, 'rb')) as wf:
            num_channels = wf.getnchannels()
            assert num_channels == 1
            sample_width = wf.getsampwidth()
            assert sample_width == 2
            sample_rate = wf.getframerate()
            assert sample_rate in (8000, 16000, 32000)
            pcm_data = wf.readframes(wf.getnframes())
            return pcm_data, sample_rate

    class Frame(object):
        """Represents a "frame" of audio data."""
        def __init__(self, bytes, timestamp, duration):
            self.bytes = bytes
            self.timestamp = timestamp
            self.duration = duration

    def _frame_generator(self, frame_duration_ms, audio, sample_rate):
        """Generates audio frames from PCM audio data.
        Takes the desired frame duration in milliseconds, the PCM data, and
        the sample rate.
        Yields Frames of the requested duration.
        """
        n = int(sample_rate * (frame_duration_ms / 1000.0) * 2)
        offset = 0
        timestamp = 0.0
        duration = (float(n) / sample_rate) / 2.0
        while offset + n < len(audio):
            yield self.Frame(audio[offset:offset + n], timestamp, duration)
            timestamp += duration
            offset += n

    def _vad_collector(self, sample_rate, frame_duration_ms,
                       padding_duration_ms, vad, frames):
        """Filters out non-voiced audio frames.
        Given a webrtcvad.Vad and a source of audio frames, yields only
        the voiced audio.
        Uses a padded, sliding window algorithm over the audio frames.
        When more than 90% of the frames in the window are voiced (as
        reported by the VAD), the collector triggers and begins yielding
        audio frames. Then the collector waits until 90% of the frames in
        the window are unvoiced to detrigger.
        The window is padded at the front and back to provide a small
        amount of silence or the beginnings/endings of speech around the
        voiced frames.
        Arguments:
        sample_rate - The audio sample rate, in Hz.
        frame_duration_ms - The frame duration in milliseconds.
        padding_duration_ms - The amount to pad the window, in milliseconds.
        vad - An instance of webrtcvad.Vad.
        frames - a source of audio frames (sequence or generator).
        Returns: A generator that yields PCM audio data.
        """
        num_padding_frames = int(padding_duration_ms / frame_duration_ms)
        # We use a deque for our sliding window/ring buffer.
        ring_buffer = collections.deque(maxlen=num_padding_frames)
        # We have two states: TRIGGERED and NOTTRIGGERED. We start in the
        # NOTTRIGGERED state.
        triggered = False

        voiced_frames = []
        for frame in frames:
            is_speech = vad.is_speech(frame.bytes, sample_rate)

            if not triggered:
                ring_buffer.append((frame, is_speech))
                num_voiced = len([f for f, speech in ring_buffer if speech])
                # If we're NOTTRIGGERED and more than 90% of the frames in
                # the ring buffer are voiced frames, then enter the
                # TRIGGERED state.
                if num_voiced > 0.9 * ring_buffer.maxlen:
                    triggered = True
                    # We want to yield all the audio we see from now until
                    # we are NOTTRIGGERED, but we have to start with the
                    # audio that's already in the ring buffer.
                    for f, s in ring_buffer:
                        voiced_frames.append(f)
                    ring_buffer.clear()
            else:
                # We're in the TRIGGERED state, so collect the audio data
                # and add it to the ring buffer.
                voiced_frames.append(frame)
                ring_buffer.append((frame, is_speech))
                num_unvoiced = len([f for f, speech in ring_buffer if not speech])
                # If more than 90% of the frames in the ring buffer are
                # unvoiced, then enter NOTTRIGGERED and yield whatever
                # audio we've collected.
                if num_unvoiced > 0.9 * ring_buffer.maxlen:
                    triggered = False
                    voiced_frames = voiced_frames[0:len(voiced_frames)-len(ring_buffer)]
                    yield voiced_frames
                    ring_buffer.clear()
                    voiced_frames = []
        if voiced_frames:
            yield voiced_frames

    def generate_missing_labels(self, lbls, start_time, end_time):

        # NOTE: wouldn't work with overlapping labels
        lbls = sorted(lbls, key=lambda x: x.start_seconds)

        missing_lbls = []

        current_time = start_time

        if lbls[0].start_seconds == start_time:
            current_time = lbls[0].end_seconds
            lbls = lbls[1:]

        for k, lbl in enumerate(lbls):
            missing_lbls.append(AudacityLabel(current_time, lbl.start_seconds, "non_speech"))
            current_time = lbl.end_seconds

        if current_time != end_time:
            missing_lbls.append(AudacityLabel(current_time, end_time, "non_speech"))

        return missing_lbls

    def detect_voice_segments(self, wav_file):
        audio, sample_rate = self._read_wave(wav_file)
        vad = webrtcvad.Vad(1)
        frames = list(self._frame_generator(30, audio, sample_rate))
        start_time = 0.0
        end_time = len(audio)/2 * 1.0/sample_rate
        voiced_frames_generator = self._vad_collector(sample_rate, 30, 150, vad, frames)
        lbls = []
        for voiced_frames in voiced_frames_generator:
            lbls.append(AudacityLabel(voiced_frames[0].timestamp, voiced_frames[-1].timestamp, "speech"))
        non_speech_lbls = self.generate_missing_labels(lbls, start_time, end_time)
        return sorted(lbls+non_speech_lbls, key=lambda x: x.start_seconds)



import logging
import shutil
import subprocess
import csv
import tempfile
import uuid
import sys
import glob

import os
from sac.model.audacity_label import AudacityLabel
from sac.util import Util


def all_same(items):
    return all(x[2] == items[0][2] for x in items)


def find_min_start(items):
    min = float('Inf')
    for i in items:
        if i[0] <= min:
            min = i[0]
    return min


def find_max_end(items):
    max = -float('Inf')
    for i in items:
        if i[1] >= max:
            max = i[1]
    return max


def chunks(l, n):
    c = []
    for i in range(0, len(l), n):
        c.append(l[i:i+n] + ["m"])
    return c


def not_in_overlapping(item, overlapping_items):
    for i in overlapping_items:
        for j in i:
            if item == j:
                return False
    return True

class WavEditor(object):

    @staticmethod
    def create_labels_from_segments(segments_dir):
        end_point = 0
        labels = []
        wavs = glob.glob(segments_dir + "/*.wav")
        for wav in wavs:
            duration = subprocess.check_output(["soxi", "-D", wav])
            start_point = end_point
            end_point = start_point + float(duration)
            segment_class = wav.split(".")[-2].split("_")[-1]
            labels.append(AudacityLabel(start_point, end_point, segment_class))

        return labels


    @staticmethod
    def create_audio_segment(start_time, end_time, input_wav, output_wav):
        duration = str(float(end_time) - float(start_time))
        subprocess.check_call(["sox", input_wav, output_wav, "trim", start_time, duration])

    @staticmethod
    def are_labels_overlapping(label1, label2):
        """
        non-overlap cases

        |---------| lbl1
                      |--------| lbl2

                    |---------| lbl1
        |---------| lbl2


        :param label1:
        :param label2:
        :return:
        """

        if (
                                label2.start_seconds > label1.end_seconds and
                                label2.start_seconds > label1.start_seconds and
                                label2.end_seconds > label1.start_seconds and
                                label2.end_seconds > label1.start_seconds) or (
                                label2.start_seconds < label1.start_seconds and
                                label2.end_seconds < label1.start_seconds and
                                label2.start_seconds < label1.end_seconds and
                                label2.end_seconds < label1.end_seconds):
            return False
        else:
            return True


    @staticmethod
    def get_non_overlapping_items_v2(rows_in_f2):

        transformed_rows = []
        for row in rows_in_f2:
            transformed_rows.append([float(row[0]), float(row[0])+float(row[1]), row[2]])

        print(transformed_rows)

        overlapping_rows = []
        rescan = True

        while rescan:
            print("rescan")
            overlapping = False
            for i in range(len(transformed_rows)):
                row1 = transformed_rows[i]
                lbl1 = AudacityLabel(row1[0], row1[1], row1[2])
                for j in range(i+1, len(transformed_rows)):
                    row2 = transformed_rows[j]
                    lbl2 = AudacityLabel(row2[0], row2[1], row2[2])

                    print(lbl1)
                    print(lbl2)

                    if WavEditor.are_labels_overlapping(lbl1, lbl2):
                        print("overlapping!")
                        overlapping_rows.append(row1)
                        overlapping_rows.append(row2)
                        overlapping = True
                        del transformed_rows[i]
                        del transformed_rows[j-1]
                        break
                if overlapping:
                    break

            if overlapping:
                rescan = True
            else:
                rescan = False

            print(transformed_rows)

        return transformed_rows




    @staticmethod
    def get_non_overlapping_items(rows_in_f2):

        # transform rows from f2 to f1
        transformed_rows = []
        for row in rows_in_f2:
            transformed_rows.append([float(row[0]), float(row[0])+float(row[1]), row[2]])

        overlapping = []
        non_overlapping = []

        for i in range(len(transformed_rows)):
            row = transformed_rows[i]
            overlapping_rows = []
            for j in range(i+1, len(transformed_rows)):
                j_row = transformed_rows[j]
                if j_row[0] < row[1]:
                    if len(overlapping_rows) == 0:
                        overlapping_rows.extend([row, j_row])
                    else:
                        overlapping_rows.append(j_row)
            if len(overlapping_rows) == 0 and not_in_overlapping(row, overlapping):
                non_overlapping.append(row)
            elif len(overlapping_rows) != 0 and not_in_overlapping(row, overlapping):
                overlapping.append(overlapping_rows)

        print(overlapping)
        print(non_overlapping)

        new_overlapping = []

        # find same class
        for o in overlapping:
            if all_same(o):
                non_overlapping.append([
                    find_min_start(o),
                    find_max_end(o),
                    o[0][2]
                ])
            else:
                new_overlapping.append(o)

        overlapping = new_overlapping

        # find m s s s .. pattern
        for o in overlapping:
            if o[0][2] == "m" and all_same(o[1:]):
                cut_points = [o[0][0]]
                m_section = o[0]
                has_end = True
                for overlapping_s_section in o[1:]:
                    if float(overlapping_s_section[0]) > float(m_section[0]) and float(overlapping_s_section[1]) < float(m_section[1]):
                        cut_points.append(overlapping_s_section[0])
                        cut_points.append(overlapping_s_section[1])
                    else:
                        cut_points.append(overlapping_s_section[0])
                        has_end = False
                if has_end:
                    cut_points.append(m_section[1])
            non_overlapping.extend(chunks(cut_points, 2))

        for i in range(len(non_overlapping)):
            non_overlapping[i][0] = str(non_overlapping[i][0])
            non_overlapping[i][1] = str(non_overlapping[i][1])
            non_overlapping[i][2] = str(non_overlapping[i][2])


        return non_overlapping

    @staticmethod
    def create_audio_segments(labels, input_wav, output_dir, clear_output_dir, delimiter, format, remove_overlapping=False):
        if clear_output_dir:
            Util.remove_dir(output_dir)
        Util.make_dir(output_dir)

        rows = WavEditor.get_rows(labels, delimiter)
        if format == "f2" and remove_overlapping:
            rows = WavEditor.get_non_overlapping_items(rows)

        for k, row in enumerate(rows):
            if format == "f1" or (format == "f2" and remove_overlapping):
                start_time = row[0]
                end_time = row[1]
            elif format == "f2":
                start_time = row[0]
                end_time = str(float(row[0]) + float(row[1]))
            else:
                logging.error("not supported file format")
                sys.exit(1)

            label = row[2]
            filename = str(uuid.uuid4())
            WavEditor.create_audio_segment(start_time, end_time, input_wav,
                                           os.path.join(output_dir, "%s_%s.wav" % (filename, label)))


    @staticmethod
    def get_rows(filename, delimiter):
        rows = []
        with open(filename, 'r') as csv_file:
            reader = csv.reader(csv_file, delimiter=delimiter)
            for row in reader:
                if len(row) > 0:
                    rows.append(row)
        return rows

    @staticmethod
    def get_files_grouped_by_class(input_dir):
        input_dir = os.path.abspath(input_dir)
        files = os.listdir(input_dir)
        files_grouped_by_class = {}

        for file in files:
            file_class = Util.get_class_from_filename(file)
            if not file_class in files_grouped_by_class:
                files_grouped_by_class[file_class] = []
            files_grouped_by_class[file_class].append(os.path.join(input_dir, file))

        return files_grouped_by_class

    @staticmethod
    def combine_audio_segments(input_dir, output_name, output_dir, clear_output_dir):
        if clear_output_dir:
            Util.remove_dir(output_dir)
        Util.make_dir(output_dir)
        files_grouped_by_class = WavEditor.get_files_grouped_by_class(input_dir)

        for class_name in files_grouped_by_class:
            input_wavs = []
            for k, file in enumerate(files_grouped_by_class[class_name]):
                input_wavs.append(file)
            cmd = ['sox'] + input_wavs + [os.path.join(output_dir, '%s_%s.wav' % (output_name, class_name))]
            subprocess.check_call(cmd)

    @staticmethod
    def split_concat(files_labels_list_file, output_name, output_dir, clear_output_dir, delimiter, format):
        files_labels_list_file = os.path.abspath(files_labels_list_file)
        files_labels_list_dir = os.path.split(files_labels_list_file)[0]

        with open(files_labels_list_file, 'rb') as f:
            files_labels_list = f.readlines()

        temp_dir = tempfile.mkdtemp()

        for file_label_pair in files_labels_list:
            label_file, audio_file = file_label_pair.decode("utf-8").strip().split(",")
            label_file_abs = os.path.join(files_labels_list_dir, label_file)
            audio_file_abs = os.path.join(files_labels_list_dir, audio_file)
            print("%s :: %s" % (label_file_abs, audio_file_abs))
            WavEditor.create_audio_segments(label_file_abs, audio_file_abs, temp_dir, clear_output_dir, delimiter,
                                            format)

        WavEditor.combine_audio_segments(temp_dir, output_name, output_dir, clear_output_dir)

        # temp_dir.cleanup()
        shutil.rmtree(temp_dir)